/*****************************************************************************
* qtcapture.m: qtkit (Mac OS X) based capture module
*****************************************************************************
* Copyright (C) 2008 the VideoLAN team
*
* Authors: Pierre d'Herbemont <pdherbemont@videolan.org>
*
*****************************************************************************
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation;
* version 2 of the License.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*****************************************************************************/

/*****************************************************************************
* Preamble
*****************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <vlc_input.h>
#include <vlc_demux.h>
#include <vlc_interface.h>
#include <vlc_dialog.h>
#include <vlc_access.h>
#include <mach/mach.h>
#include <mach/mach_time.h>
#include <CoreServices/CoreServices.h>

#import <QTKit/QTKit.h>
#import <CoreAudio/CoreAudio.h>


#define QTKIT_WIDTH_TEXT N_("Video Capture width")
#define QTKIT_WIDTH_LONGTEXT N_("Video Capture width in pixel")
#define QTKIT_HEIGHT_TEXT N_("Video Capture height")
#define QTKIT_HEIGHT_LONGTEXT N_("Video Capture height in pixel")

/*****************************************************************************
* Local prototypes
*****************************************************************************/
static int Open( vlc_object_t *p_this );
static void Close( vlc_object_t *p_this );
static int Demux( demux_t *p_demux );
static int Control( demux_t *, int, va_list );

/*****************************************************************************
* Module descriptor
*****************************************************************************/
vlc_module_begin ()
   set_shortname( N_("Quicktime Capture") )
   set_description( N_("Quicktime Capture") )
   set_category( CAT_INPUT )
   set_subcategory( SUBCAT_INPUT_ACCESS )
   add_shortcut( "qtcapture" )
   set_capability( "access_demux", 10 )
   set_callbacks( Open, Close )
   add_integer("qtcapture-width", 1024, NULL, QTKIT_WIDTH_TEXT, QTKIT_WIDTH_LONGTEXT, true)
    change_integer_range (80, 1280)
   add_integer("qtcapture-height", 768, NULL, QTKIT_HEIGHT_TEXT, QTKIT_HEIGHT_LONGTEXT, true)
	change_integer_range (60, 800)
vlc_module_end ()


/*****************************************************************************
* QTKit Bridge
*****************************************************************************/
@interface VLCDecompressedVideoOutput : QTCaptureDecompressedVideoOutput
{
    CVImageBufferRef currentImageBuffer;
    mtime_t currentPts;
    mtime_t previousPts;
}
- (id)init;
- (void)outputVideoFrame:(CVImageBufferRef)videoFrame withSampleBuffer:(QTSampleBuffer *)sampleBuffer fromConnection:(QTCaptureConnection *)connection;
- (mtime_t)copyCurrentFrameToBuffer:(void *)buffer;
@end

/* Apple sample code */
@implementation VLCDecompressedVideoOutput : QTCaptureDecompressedVideoOutput
- (id)init
{
    if( self = [super init] )
    {
        currentImageBuffer = nil;
        currentPts = 0;
        previousPts = 0;
    }
    return self;
}
- (void)dealloc
{
    @synchronized (self)
    {
        CVBufferRelease(currentImageBuffer);
        currentImageBuffer = nil;
    }
    [super dealloc];
}

- (void)outputVideoFrame:(CVImageBufferRef)videoFrame withSampleBuffer:(QTSampleBuffer *)sampleBuffer fromConnection:(QTCaptureConnection *)connection
{
    // Store the latest frame
    // This must be done in a @synchronized block because this delegate method is not called on the main thread
    CVImageBufferRef imageBufferToRelease;

    CVBufferRetain(videoFrame);

    @synchronized (self)
    {
        imageBufferToRelease = currentImageBuffer;
        currentImageBuffer = videoFrame;
        currentPts = (mtime_t)(1000000L / [sampleBuffer presentationTime].timeScale * [sampleBuffer presentationTime].timeValue);
        
        /* Try to use hosttime of the sample if available, because iSight Pts seems broken */
        NSNumber *hosttime = (NSNumber *)[sampleBuffer attributeForKey:QTSampleBufferHostTimeAttribute];
        //if( hosttime ) currentPts = (mtime_t)AudioConvertHostTimeToNanos([hosttime unsignedLongLongValue])/1000;
		NSLog(@"video hosttime: %u",hosttime);
		NSLog(@"video currentPts: %u",currentPts);
    }
    CVBufferRelease(imageBufferToRelease);
}

- (mtime_t)copyCurrentFrameToBuffer:(void *)buffer
{
    CVImageBufferRef imageBuffer;
    mtime_t pts;
	void * pixels;

    if(!currentImageBuffer || currentPts == previousPts )
        return 0;

    @synchronized (self)
    {
        imageBuffer = CVBufferRetain(currentImageBuffer);
		if(imageBuffer){
		            pts = previousPts = currentPts;
		            CVPixelBufferLockBaseAddress(imageBuffer, 0);
		            pixels = CVPixelBufferGetBaseAddress(imageBuffer);
		            if(pixels)
		                memcpy( buffer, pixels, CVPixelBufferGetBytesPerRow(imageBuffer) * CVPixelBufferGetHeight(imageBuffer));
		            CVPixelBufferUnlockBaseAddress(imageBuffer, 0);
		        }
    }
	
    CVBufferRelease(imageBuffer);

    if(pixels)
			return currentPts;
	else
			return 0;
}

@end


@interface VLCDecompressedAudioOutput : QTCaptureDecompressedAudioOutput
{
    AudioBuffer *currentAudioBuffer;
    mtime_t currentPts;
	mtime_t mach_currentPts;
    mtime_t previousPts;
}
- (id)init;
- (void)outputAudioSampleBuffer:(QTSampleBuffer *)sampleBuffer fromConnection:(QTCaptureConnection *)connection;
- (mtime_t)copyCurrentAudioBytesToBuffer:(void *)buffer;
@end

@implementation VLCDecompressedAudioOutput : QTCaptureDecompressedAudioOutput
- (id)init
{
    if( self = [super init] )
    {
        currentAudioBuffer = nil;
        currentPts = 0;
        previousPts = 0;
    }
    return self;
}
- (void)dealloc
{
    @synchronized (self)
    {
        currentAudioBuffer = nil;
    }
    [super dealloc];
}

- (void)outputAudioSampleBuffer:(QTSampleBuffer *)sampleBuffer fromConnection:(QTCaptureConnection *)connection
{
	//NSLog(@"DEBUG outputAudioSampleBuffer:");
	AudioBufferList *tempAudioBufferList;
	NSMutableData *rawAudioData;
	AudioSampleType *audioData;
	//AudioStreamBasicDescription sampleBufferASBD;
	int audioDataSize = 0;
	int count = 0;
	
	@synchronized (self)
    {
        //audioBufferToRelease = malloc(sizeof(currentAudioBufferList));
		//audioBufferToRelease = currentAudioBufferList;
		tempAudioBufferList = [sampleBuffer audioBufferListWithOptions:QTSampleBufferAudioBufferListOptionAssure16ByteAlignment];
		QTFormatDescription *formatDescription = [sampleBuffer formatDescription];
		NSValue *sampleBufferASBDValue = [formatDescription attributeForKey:QTFormatDescriptionAudioStreamBasicDescriptionAttribute];
		if (!sampleBufferASBDValue)
		{
			return;
		}
		
		AudioStreamBasicDescription sampleBufferASBD = {0};
		[sampleBufferASBDValue getValue:&sampleBufferASBD];
		
		char formatIDString[5];
		UInt32 formatID = CFSwapInt32HostToBig (sampleBufferASBD.mFormatID);
		bcopy (&formatID, formatIDString, 4);
		formatIDString[4] = '\0';
		
		NSLog (@"  Sample Rate:         %10.0f",  sampleBufferASBD.mSampleRate);
		NSLog (@"  Format ID:           %10s",    formatIDString);
		NSLog (@"  Format Flags:        %10X",    sampleBufferASBD.mFormatFlags);
		NSLog (@"  Bytes per Packet:    %10d",    sampleBufferASBD.mBytesPerPacket);
		NSLog (@"  Frames per Packet:   %10d",    sampleBufferASBD.mFramesPerPacket);
		NSLog (@"  Bytes per Frame:     %10d",    sampleBufferASBD.mBytesPerFrame);
		NSLog (@"  Channels per Frame:  %10d",    sampleBufferASBD.mChannelsPerFrame);
		NSLog (@"  Bits per Channel:    %10d",    sampleBufferASBD.mBitsPerChannel);
		
		audioData = (AudioSampleType *) calloc (sampleBufferASBD.mFramesPerPacket, sizeof (AudioSampleType));
		
		currentPts = (mtime_t)(1000000L / [sampleBuffer presentationTime].timeScale * [sampleBuffer presentationTime].timeValue);
        //mach_currentPts = (mtime_t)mach_absolute_time();
		// Try to use hosttime of the sample if available, as with video 
        NSNumber *hosttime = (NSNumber *)[sampleBuffer attributeForKey:QTSampleBufferHostTimeAttribute];
        if( hosttime ) currentPts = (mtime_t)AudioConvertHostTimeToNanos([hosttime unsignedLongLongValue])/1000;
		
		// Time debugging
		//NSLog(@"currentPts %d",currentPts);
		//NSLog(@"\noriginalPts: %@;\ntimeScale: %d;\ntimeValue: %d;\ntimeScale*timeValue: %lld;\nvlc_calculation: %d;\nhosttime: %i;\nmach_currentPts: %i",QTStringFromTime([sampleBuffer presentationTime]),[sampleBuffer presentationTime].timeScale,[sampleBuffer presentationTime].timeValue,([sampleBuffer presentationTime].timeScale * [sampleBuffer presentationTime].timeValue),(1000000L / [sampleBuffer presentationTime].timeScale * [sampleBuffer presentationTime].timeValue),[[sampleBuffer attributeForKey:QTSampleBufferHostTimeAttribute] intValue],mach_currentPts);
		//NSLog(@"currentAudioBufferList: %d", tempAudioBufferList->mNumberBuffers);
		NSLog(@"audio hosttime: %u",hosttime);
		NSLog(@"audio currentPts: %u",currentPts);
		
		if (tempAudioBufferList->mNumberBuffers > 1)
		{
			for (count = 0; count < tempAudioBufferList->mNumberBuffers; count++) {
				audioDataSize = (audioDataSize + tempAudioBufferList->mBuffers[count].mDataByteSize);
				
			}
			//NSLog(@"audioDataSize: %d",audioDataSize);
			rawAudioData = [[NSMutableData alloc] initWithLength:audioDataSize];
		}
		
		for (count = 0; count < tempAudioBufferList->mNumberBuffers; count++) {
			//NSLog(@"count: %d; channels: %d; size %d; data: %d",count,tempAudioBufferList->mBuffers[count].mNumberChannels,tempAudioBufferList->mBuffers[count].mDataByteSize,tempAudioBufferList->mBuffers[count].mData);
			//[rawAudioData appendBytes:tempAudioBufferList->mBuffers[count].mData length:tempAudioBufferList->mBuffers[count].mDataByteSize];			
			audioData = tempAudioBufferList->mBuffers[count].mData;
		}
		currentAudioBuffer = (AudioBuffer *)malloc(sizeof(AudioBuffer));
		if (NULL == currentAudioBuffer) {
			//NSLog (@"*** malloc failure for allocating audioBuffer memory");
			[rawAudioData release];
			return;
		}
		
		//currentAudioBuffer->mNumberChannels = tempAudioBufferList->mBuffers[0].mNumberChannels;
		currentAudioBuffer->mNumberChannels = 1;
		currentAudioBuffer->mDataByteSize = sizeof(AudioSampleType);
		currentAudioBuffer->mData = audioData;
		//currentAudioBuffer->mData = tempAudioBufferList->mBuffers[0].mData;
		//NSLog(@"currentAudioBuffer channels: %d; size: %d; data: %d",currentAudioBuffer->mNumberChannels,currentAudioBuffer->mDataByteSize,currentAudioBuffer->mData);
    }
	//NSLog(@"audio currentPts %d",currentPts);
	tempAudioBufferList = nil;
	[rawAudioData release];
	//free(audioBufferToRelease);
	//[audioBufferToRelease release];
}

- (mtime_t)copyCurrentAudioBytesToBuffer:(void *)buffer
{
	//NSLog(@"DEBUG copyCurrentAudioBytesToBuffer:");
	AudioBuffer *audioBuffer;
    mtime_t pts;
	//void * data;

    if(!currentAudioBuffer|| currentPts == previousPts )
	{
		//NSLog(@"nothing there");
		return 0;
	}

    @synchronized (self)
    {
		audioBuffer = (AudioBuffer *)malloc(sizeof(currentAudioBuffer));
		if(audioBuffer){
			//NSLog(@"currentAudioBuffer channels: %d; size: %d; data: %d",currentAudioBuffer->mNumberChannels,currentAudioBuffer->mDataByteSize,currentAudioBuffer->mData);
			pts = previousPts = currentPts;
			memcpy( buffer, currentAudioBuffer->mData, currentAudioBuffer->mDataByteSize);
		}
        //audioBuffer = (AudioBufferList *)malloc(sizeof(AudioBufferList));
		//audioBuffer = malloc(sizeof(currentAudioBufferList));
		//NSLog(@"audiobuffer");
		
		/*
		if( !(currentAudioBufferList->mBuffers[0].mData) || (currentAudioBufferList->mNumberBuffers == 0) )
		{
			NSLog(@"No data yet");
			return 0;
		}
		*/
		
		
		/*
		NSMutableData *mergedAudioBufferData = [NSMutableData dataWithCapacity:sizeof(AudioBufferList)];
		int dataIndex;
		for (dataIndex = 0; dataIndex < currentAudioBufferList->mNumberBuffers; dataIndex++) {
			NSLog(@"merge");
			//NSLog(@"merging buffer: %@ into one buffer with data: ",currentAudioBufferList->mNumberBuffers, currentAudioBufferList->mBuffers[dataIndex].mData);
			[mergedAudioBufferData appendBytes:currentAudioBufferList->mBuffers[dataIndex].mData length:currentAudioBufferList->mBuffers[dataIndex].mDataByteSize];
		}
		
		NSLog(@"creating new AudioBuffer");
		audioBuffer->mNumberBuffers = 1;
		audioBuffer->mBuffers[0].mNumberChannels = currentAudioBufferList->mBuffers[0].mNumberChannels;
		audioBuffer->mBuffers[0].mDataByteSize = [mergedAudioBufferData length];
		audioBuffer->mBuffers[0].mData = [mergedAudioBufferData mutableBytes];
		*/
		//pts = previousPts = currentPts;
		
		//data = [mergedAudioBufferData mutableBytes];
		//memcpy( buffer , data, 0);
		//[mergedAudioBufferData release];
		//[data release];
    }
	//free(audioBuffer);
	free(audioBuffer);
	
	if(currentAudioBuffer->mData)
		return currentPts;
	else
		return 0;
}

@end

 
/*****************************************************************************
* Struct
*****************************************************************************/

struct demux_sys_t {
    QTCaptureSession * session;
    QTCaptureDevice * device;
	QTCaptureDevice * audiodevice;
    VLCDecompressedVideoOutput * output;
	VLCDecompressedAudioOutput * audiooutput;
    int height, width;
	int i_audio_max_frame_size;
	int channels, i_sample_rate;
    es_out_id_t * p_es_video;
	es_out_id_t * p_es_audio;
};


/*****************************************************************************
* qtchroma_to_fourcc
*****************************************************************************/
static int qtchroma_to_fourcc( int i_qt )
{
    static const struct
    {
        unsigned int i_qt;
        int i_fourcc;
    } qtchroma_to_fourcc[] =
    {
        /* Raw data types */
        { '2vuy',    VLC_CODEC_UYVY },
        { 'yuv2',VLC_CODEC_YUYV },
        { 'yuvs', VLC_CODEC_YUYV },
        { 0, 0 }
    };
    int i;
    for( i = 0; qtchroma_to_fourcc[i].i_qt; i++ )
    {
        if( qtchroma_to_fourcc[i].i_qt == i_qt )
            return qtchroma_to_fourcc[i].i_fourcc;
    }
    return 0;
}

/*****************************************************************************
* Open:
*****************************************************************************/
static int Open( vlc_object_t *p_this )
{
    demux_t     *p_demux = (demux_t*)p_this;
    demux_sys_t *p_sys = NULL;
    es_format_t fmt;
	es_format_t audiofmt;
    int i;
    int i_width;
    int i_height;
    int result = 0;
	char *psz_uid = NULL;

    /* Only when selected */
    if( *p_demux->psz_access == '\0' )
        return VLC_EGENERIC;
    
	if(p_demux->psz_path && *p_demux->psz_path ){
        psz_uid = strdup(p_demux->psz_path);
    }
    msg_Dbg( p_demux, "qtcapture uid = %s", psz_uid );
    NSString *qtk_currdevice_uid = [[NSString alloc] initWithFormat:@"%s", psz_uid];
	//MIFEURST begin
	NSString *qtk_curraudiodevice_uid = [[NSString alloc] initWithString:@"iMic USB audio system"];
	//NSString *qtk_curraudiodevice_uid = [[NSString alloc] initWithString:@"Built-in Microphone"];
	//MIFEURST end
	
	
    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];

    /* Set up p_demux */
    p_demux->pf_demux = Demux;
    p_demux->pf_control = Control;
    p_demux->info.i_update = 0;
    p_demux->info.i_title = 0;
    p_demux->info.i_seekpoint = 0;
    
    p_demux->p_sys = p_sys = calloc( 1, sizeof( demux_sys_t ) );
    if( !p_sys )
        return VLC_ENOMEM;
    
    msg_Dbg( p_demux, "qtcapture : uid = %s", [qtk_currdevice_uid UTF8String]);
    NSArray *myVideoDevices = [[[QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeVideo] arrayByAddingObjectsFromArray:[QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeMuxed]] retain];
    if([myVideoDevices count] == 0)
    {
        dialog_FatalWait( p_demux, _("No Input device found"),
                        _("Your Mac does not seem to be equipped with a suitable input device. "
                          "Please check your connectors and drivers.") );
        msg_Err( p_demux, "Can't find any Video device" );
        
        goto error;
    }
    int ivideo;
    for(ivideo = 0; ivideo < [myVideoDevices count]; ivideo++){
        QTCaptureDevice *qtk_device;
        qtk_device = [myVideoDevices objectAtIndex:ivideo];
        msg_Dbg( p_demux, "qtcapture video %d/%d %s %s", ivideo, [myVideoDevices count], [[qtk_device localizedDisplayName] UTF8String], [[qtk_device uniqueID] UTF8String]);
        if([[[qtk_device uniqueID]stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]] isEqualToString:qtk_currdevice_uid]){
            break;
        }
    } 
	
	//MIFEURST begin
	NSArray *myAudioDevices = [[[QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeSound] arrayByAddingObjectsFromArray:[QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeMuxed]] retain];
    if([myAudioDevices count] == 0)
    {
        dialog_FatalWait( p_demux, _("No Audio Input device found"),
						 _("Your Mac does not seem to be equipped with a suitable audio input device. "
						   "Please check your connectors and drivers.") );
        msg_Err( p_demux, "Can't find any Audio device" );
        
        goto error;
    }
    int iaudio;
    for(iaudio = 0; iaudio < [myAudioDevices count]; iaudio++){
        QTCaptureDevice *qtk_audioDevice;
        qtk_audioDevice = [myAudioDevices objectAtIndex:iaudio];
        msg_Dbg( p_demux, "qtcapture audio %d/%d localizedDisplayName: %s uniqueID: %s", iaudio, [myAudioDevices count], [[qtk_audioDevice localizedDisplayName] UTF8String], [[qtk_audioDevice uniqueID] UTF8String]);
        if([[[qtk_audioDevice localizedDisplayName]stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]] isEqualToString:qtk_curraudiodevice_uid]){
            msg_Dbg( p_demux, "Device found" );
			break;
        }
    }
	//MIFEURST end
	
    memset( &fmt, 0, sizeof( es_format_t ) );    
    
    QTCaptureDeviceInput * input = nil;
    NSError *o_returnedError;

    if(ivideo < [myVideoDevices count]){
		p_sys->device = [myVideoDevices objectAtIndex:ivideo];
	}
	else
    {
        /* cannot find designated device, fall back to open default device */
        msg_Dbg(p_demux, "Cannot find designated uid device as %s. Fall back to open default.", [qtk_currdevice_uid UTF8String]);
        p_sys->device = [QTCaptureDevice defaultInputDeviceWithMediaType: QTMediaTypeVideo];
    }
    if( !p_sys->device )
    {
        dialog_FatalWait( p_demux, _("No Input device found"),
                        _("Your Mac does not seem to be equipped with a suitable input device. "
                          "Please check your connectors and drivers.") );
        msg_Err( p_demux, "Can't find any Video device" );
        
        goto error;
    }

    if( ![p_sys->device open: &o_returnedError] )
    {
        msg_Err( p_demux, "Unable to open the capture device (%ld)", [o_returnedError code] );
        goto error;
    }

    if( [p_sys->device isInUseByAnotherApplication] == YES )
    {
        msg_Err( p_demux, "default capture device is exclusively in use by another application" );
        goto error;
    }

    input = [[QTCaptureDeviceInput alloc] initWithDevice: p_sys->device];
    if( !input )
    {
        msg_Err( p_demux, "can't create a valid capture input facility" );
        goto error;
    }
	
	//MIFEURST begin
	memset( &audiofmt, 0, sizeof( es_format_t ) ); 
	
	QTCaptureDeviceInput * audioInput = nil;
    NSError *o_returnedAudioError;
    if(ivideo < [myAudioDevices count]){
		p_sys->audiodevice = [myAudioDevices objectAtIndex:iaudio];
	}
	else
    {
        /* cannot find designated audio device, fall back to open default audio device */
        msg_Dbg(p_demux, "Cannot find designated uid audio device as %s. Fall back to open default audio device.", [qtk_curraudiodevice_uid UTF8String]);
        p_sys->audiodevice = [QTCaptureDevice defaultInputDeviceWithMediaType: QTMediaTypeSound];
    }
    if( !p_sys->audiodevice )
    {
        dialog_FatalWait( p_demux, _("No audio input device found"),
						 _("Your Mac does not seem to be equipped with a suitable audio input device. "
						   "Please check your connectors and drivers.") );
        msg_Err( p_demux, "Can't find any Audio device" );
        
        goto error;
    }
	
    if( ![p_sys->audiodevice open: &o_returnedAudioError] )
    {
        msg_Err( p_demux, "Unable to open the audio capture device (%ld)", [o_returnedAudioError code] );
        goto error;
    }
	
    if( [p_sys->audiodevice isInUseByAnotherApplication] == YES )
    {
        msg_Err( p_demux, "default audio capture device is exclusively in use by another application" );
        goto error;
    }
    audioInput = [[QTCaptureDeviceInput alloc] initWithDevice: p_sys->audiodevice];
    if( !audioInput )
    {
        msg_Err( p_demux, "can't create a valid audio capture input facility" );
        goto error;
    } else {
		msg_Dbg( p_demux, "created valid audio capture input facility" );
	}

	//MIFEURST end

    p_sys->output = [[VLCDecompressedVideoOutput alloc] init];
	
	//MIFEURST begin
	p_sys->audiooutput = [[VLCDecompressedAudioOutput alloc] init];
	msg_Dbg ( p_demux, "initialized video and audio output" );
	//MIFEURST end

    /* Get the formats */
    NSArray *format_array = [p_sys->device formatDescriptions];
    QTFormatDescription* camera_format = NULL;
    for( int k = 0; k < [format_array count]; k++ )
    {
        camera_format = [format_array objectAtIndex: k];

		msg_Dbg( p_demux, "Video localized format summary: %s", [[camera_format localizedFormatSummary] UTF8String]);
		msg_Dbg( p_demux, "Video format description attributes: %s",[[[camera_format formatDescriptionAttributes] description] UTF8String]);
    }
    if( [format_array count] )
        camera_format = [format_array objectAtIndex: 0];
    else goto error;
	
	//MIFEURST begin
	NSArray *audioformat_array = [p_sys->audiodevice formatDescriptions];
    QTFormatDescription* audio_format = NULL;
    for( int k = 0; k < [audioformat_array count]; k++ )
    {
        audio_format = (QTFormatDescription *)[audioformat_array objectAtIndex: k];
		
		msg_Dbg( p_demux, "Audio localized format summary: %s", [[audio_format localizedFormatSummary] UTF8String]);
		msg_Dbg( p_demux, "Audio format description attributes: %s",[[[audio_format formatDescriptionAttributes] description] UTF8String]);	
		
		// TODO: figure out a way to read from the ASBD
		AudioStreamBasicDescription audioDesc = {0};
		memset (&audioDesc, 0, sizeof (audioDesc));
		NSValue *descDataValue =  [audio_format attributeForKey:QTFormatDescriptionAudioStreamBasicDescriptionAttribute];
		[descDataValue getValue:&audioDesc];
		
		msg_Dbg( p_demux, "Audio format detail: %i",audioDesc );

    }
    if( [audioformat_array count] )
        audio_format = [format_array objectAtIndex: 0];
    else goto error;
	//MIFEURST end

    int qtchroma = [camera_format formatType];
	int chroma = VLC_CODEC_UYVY;

    /* Now we can init */
    es_format_Init( &fmt, VIDEO_ES, chroma );

    NSSize encoded_size = [[camera_format attributeForKey:QTFormatDescriptionVideoEncodedPixelsSizeAttribute] sizeValue];
    NSSize display_size = [[camera_format attributeForKey:QTFormatDescriptionVideoCleanApertureDisplaySizeAttribute] sizeValue];
    NSSize par_size = [[camera_format attributeForKey:QTFormatDescriptionVideoProductionApertureDisplaySizeAttribute] sizeValue];

    par_size.width = display_size.width = encoded_size.width
        = var_CreateGetInteger (p_this, "qtcapture-width");
    par_size.height = display_size.height = encoded_size.height
        = var_CreateGetInteger (p_this, "qtcapture-height");
	
    fmt.video.i_width = p_sys->width = encoded_size.width;
    fmt.video.i_height = p_sys->height = encoded_size.height;
    if( par_size.width != encoded_size.width )
    {
        fmt.video.i_sar_num = (int64_t)encoded_size.height * par_size.width / encoded_size.width;
        fmt.video.i_sar_den = encoded_size.width;
    }

    msg_Dbg(p_demux, "encoded_size %d %d", (int)encoded_size.width, (int)encoded_size.height );
    msg_Dbg(p_demux, "display_size %d %d", (int)display_size.width, (int)display_size.height );
    msg_Dbg(p_demux, "PAR size %d %d", (int)par_size.width, (int)par_size.height );
    
    [p_sys->output setPixelBufferAttributes: [NSDictionary dictionaryWithObjectsAndKeys:
        [NSNumber numberWithUnsignedInt:kCVPixelFormatType_422YpCbCr8], (id)kCVPixelBufferPixelFormatTypeKey,
		[NSNumber numberWithInt: p_sys->height], kCVPixelBufferHeightKey,
        [NSNumber numberWithInt: p_sys->width], kCVPixelBufferWidthKey,
        [NSNumber numberWithBool:YES], (id)kCVPixelBufferOpenGLCompatibilityKey,
        nil]];
	
	//MIFEURST begin
	//int audiocodec = VLC_CODEC_S16B;
	int audiocodec = VLC_CODEC_ALAW;
	//int audiocodec = VLC_CODEC_DVDA_LPCM;
	//int audiocodec = VLC_FOURCC('s','o','w','t');
	
	es_format_Init( &audiofmt, AUDIO_ES, audiocodec);
	
	//AudioStreamBasicDescription audio_format_ASBDValue = [audio_format attributeForKey:QTFormatDescriptionAudioStreamBasicDescriptionAttribute];
	//p_sys->i_audio_max_frame_size = sizeof(audio_format_ASBDValue);
	
	// FIXME: these values should be set dynamically
	p_sys->i_audio_max_frame_size = 32 * 1024;
	p_sys->channels = 2;
	p_sys->i_sample_rate = 44100;
	audiofmt.audio.i_channels = 2;
	audiofmt.audio.i_rate = 44100;
	audiofmt.audio.i_bitspersample = 16;
	audiofmt.audio.i_blockalign = audiofmt.audio.i_channels * audiofmt.audio.i_bitspersample / 8;
	audiofmt.i_bitrate = audiofmt.audio.i_channels * audiofmt.audio.i_rate * audiofmt.audio.i_bitspersample;
	
	//MIFEURST end

    p_sys->session = [[QTCaptureSession alloc] init];

	// TODO: minimize those bools
    bool retVideo = [p_sys->session addInput:input error: &o_returnedError];
    if( !retVideo )
    {
        msg_Err( p_demux, "the video capture device could not be added to capture session (%ld)", [o_returnedError code] );
        goto error;
    }
	
	bool retAudio = [p_sys->session addInput:audioInput error: &o_returnedAudioError];
    if( !retAudio )
    {
        msg_Err( p_demux, "the audio capture device could not be added to capture session (%ld)", [o_returnedAudioError code] );
        goto error;
    }

    bool retVideoOutput = [p_sys->session addOutput:p_sys->output error: &o_returnedError];
    if( !retVideoOutput )
    {
        msg_Err( p_demux, "video output could not be added to capture session (%ld)", [o_returnedError code] );
        goto error;
    }
	
	bool retAudioOutput = [p_sys->session addOutput:p_sys->audiooutput error: &o_returnedError];
    if( !retAudioOutput )
    {
        msg_Err( p_demux, "audio output could not be added to capture session (%ld)", [o_returnedError code] );
        goto error;
    }

    [p_sys->session startRunning];

    msg_Dbg( p_demux, "New video es %4.4s %dx%d",
            (char*)&fmt.i_codec, fmt.video.i_width, fmt.video.i_height );
	msg_Dbg( p_demux, "New audio es %d channels %dHz",
            audiofmt.audio.i_channels, audiofmt.audio.i_rate );

    p_sys->p_es_video = es_out_Add( p_demux->out, &fmt );
	p_sys->p_es_audio = es_out_Add( p_demux->out, &audiofmt );

    [input release];
    [pool release];

    msg_Dbg( p_demux, "QTCapture: We have a video and audio device ready!" );

    return VLC_SUCCESS;
error:
    [input release];
    [pool release];

    free( p_sys );

    return VLC_EGENERIC;
}

/*****************************************************************************
* Close:
*****************************************************************************/
static void Close( vlc_object_t *p_this )
{
    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];

    demux_t     *p_demux = (demux_t*)p_this;
    demux_sys_t *p_sys = p_demux->p_sys;

    /* Hack: if libvlc was killed, main interface thread was,
     * and poor QTKit needs it, so don't tell him.
     * Else we dead lock. */
    if( vlc_object_alive(p_this->p_libvlc))
    {
        // Perform this on main thread, as the framework itself will sometimes try to synchronously
        // work on main thread. And this will create a dead lock.
        [p_sys->session performSelectorOnMainThread:@selector(stopRunning) withObject:nil waitUntilDone:NO];
        [p_sys->output performSelectorOnMainThread:@selector(release) withObject:nil waitUntilDone:NO];
		[p_sys->audiooutput performSelectorOnMainThread:@selector(release) withObject:nil waitUntilDone:NO];
        [p_sys->session performSelectorOnMainThread:@selector(release) withObject:nil waitUntilDone:NO];
    }
    free( p_sys );

    [pool release];
}


/*****************************************************************************
* Demux:
*****************************************************************************/
static int Demux( demux_t *p_demux )
{
    demux_sys_t *p_sys = p_demux->p_sys;
    block_t *p_block_video;
	block_t *p_block_audio;
	int i_read = 0;
	int i_correct = 0;
	
	/*
	p_block_video = block_New( p_demux, p_sys->width *
							  p_sys->height * 2 );
    if( !p_block_video )
    {
        msg_Err( p_demux, "cannot get video block" );
        return 0;
    }
	 */
	
	//MIFEURST begin
	p_block_audio = block_New( p_demux, p_sys->i_audio_max_frame_size );
	
	if( !p_block_audio )
    {
        msg_Err( p_demux, "cannot get audio block" );
        return 0;
    }
	
	/*
	i_read = read( 1, p_block_audio->p_buffer,
				  p_sys->i_audio_max_frame_size );
	
	if( i_read <= 0 )
    {
        block_Release( p_block_audio );
        return NULL;
    }
	
	p_block_audio->i_buffer = i_read;
	i_correct = i_read;	
	*/
	//MIFEURST end

    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
	
	/*
    @synchronized (p_sys->output)
    {
		p_block_video->i_pts = [p_sys->output copyCurrentFrameToBuffer: p_block_video->p_buffer];
    }
	 */
	
    @synchronized (p_sys->audiooutput)
    {
		p_block_audio->i_pts = [p_sys->audiooutput copyCurrentAudioBytesToBuffer: p_block_audio->p_buffer];
		//p_block_audio->i_pts = p_block_audio->i_dts = mdate() - INT64_C(1000000) * (mtime_t)i_correct /	2 / p_sys->channels / p_sys->i_sample_rate;
    }
	
	/*
    if( !p_block_video->i_pts )
    {
        // Nothing to display yet, just forget
        block_Release( p_block_video );
        [pool release];
        msleep( 10000 );
        return 1;
    }
	 */
	
	if( !p_block_audio->i_pts )
    {
        // Nothing to transfer yet, just forget
        block_Release( p_block_audio );
        [pool release];
        msleep( 10000 );
        return 1;
    }

    //es_out_Control( p_demux->out, ES_OUT_SET_PCR, p_block_video->i_pts );
    //es_out_Send( p_demux->out, p_sys->p_es_video, p_block_video );
	
	es_out_Control( p_demux->out, ES_OUT_SET_PCR, p_block_audio->i_pts );
    es_out_Send( p_demux->out, p_sys->p_es_audio, p_block_audio );
	
	[pool release];
	
	/*
	//MIFEURST begin
	NSAutoreleasePool * pool2 = [[NSAutoreleasePool alloc] init];
	
    @synchronized (p_sys->audiooutput)
    {
		p_block_audio->i_pts = [p_sys->audiooutput copyCurrentAudioBytesToBuffer: p_block_audio->p_buffer];
    }
	
    if( !p_block_audio->i_pts )
    {
        // Nothing to transfer yet, just forget
        block_Release( p_block_audio );
        [pool2 release];
        msleep( 10000 );
        return 1;
    }
	
    es_out_Control( p_demux->out, ES_OUT_SET_PCR, p_block_audio->i_pts );
    es_out_Send( p_demux->out, p_sys->p_es_audio, p_block_audio );
	
	[pool2 release];
    [pool release];
	//MIFEURST end
	 */
	
    return 1;
}

/*****************************************************************************
* Control:
*****************************************************************************/
static int Control( demux_t *p_demux, int i_query, va_list args )
{
    bool *pb;
    int64_t    *pi64;

    switch( i_query )
    {
        /* Special for access_demux */
        case DEMUX_CAN_PAUSE:
        case DEMUX_CAN_SEEK:
        case DEMUX_SET_PAUSE_STATE:
        case DEMUX_CAN_CONTROL_PACE:
           pb = (bool*)va_arg( args, bool * );
           *pb = false;
           return VLC_SUCCESS;

        case DEMUX_GET_PTS_DELAY:
           pi64 = (int64_t*)va_arg( args, int64_t * );
           *pi64 = (int64_t)DEFAULT_PTS_DELAY;
           return VLC_SUCCESS;

        default:
           return VLC_EGENERIC;
    }
    return VLC_EGENERIC;
}
